import base64
import logging
import os

from collections import namedtuple
from dataclasses import dataclass
from typing import List, Dict, Union, Tuple, Optional

from .api import event, task_event
from .appliance import min_ppa_version, retired
from .exceptions import EventNotConfigured, EventTypeDisabled, EventConfigurationError, NoMailRecipients

logger = logging.getLogger(__name__)


class EventType(namedtuple("EventType", ("protocol", "event_name"))):
    """ An event type supported by the PPA appliance.

    :param protocol: The event protocol.
    :param event_name: A friendly name for the event used in exception messages if an error occurs.
    """


class Attachment(namedtuple("Attachment", ("filename", "path"))):
    """ An attachment to be attached to an email sent via the send_email function.

    :param filename: The file name.
    :param path: The path to the file.
    """


@dataclass(frozen=True)
class Events:
    SMTP = EventType(protocol="SMTP", event_name="email")
    CEF = EventType(protocol="syslog", event_name="CEF event")


@dataclass(frozen=True)
class CEFEventLevels:
    INFO = "info"
    PANIC = "panic"
    FATAL = "fatal"
    ERROR = "error"
    WARNING = "warning"
    DEBUG = "debug"
    TRACE = "trace"


@min_ppa_version("2.7.0")
@task_event({204: EventNotConfigured(Events.SMTP), 405: EventTypeDisabled(Events.SMTP), 422: NoMailRecipients()})
def send_email(
    *,
    subject: str,
    html: str,
    recipients: Optional[List[str]] = None,
    plain_text: Optional[str] = "Please enable HTML to view this email.",
    reason: Optional[str] = None,
    attachments: Optional[List[Attachment]] = None,
) -> None:
    """ Send an email using the PPA syslog configuration.

    :param subject: Email subject.
    :param html: The email content as a string in HTML format.
    :param recipients: An optional list of email addresses. If no recipients are provided, PPA will use the default
        recipient configured in the PPA appliance.
    :param plain_text: An optional string containing a plain-text version of the email. Useful for email clients that
        don't support HTML emails.
    :param reason: The text to show in the PPA task interface when the email is sent. If it is not supplied, no text
        will be shown.
    :param attachments: An optional list containing :class:`Attachment`.

    :raises exceptions.EventNotConfigured: If SMTP is not configured in the PPA appliance.
    :raises exceptions.EventTypeDisabled: If the PPA appliance SMTP configuration is disabled.
    :raises exceptions.NoMailRecipients: If no recipients were supplied to the function, and there are no default
        recipients in the PPA appliance SMTP configuration.

    Example:

        .. code-block:: python

            events.send_email(
                subject="An Active Directory user was moved by PPA",
                html=html_body,
                recipients=["example.recipient@email.com],
                reason="Alerting operations of Active Directory update"
            )
    """
    return event(
        "message/smtp",
        reason,
        subject=subject,
        recipients=recipients,
        html=html,
        message=plain_text,
        files=[("file", (attachment.filename, open(attachment.path, "rb"))) for attachment in attachments]
        if attachments
        else None,
    )


@min_ppa_version("2.4.0")
@retired("2.5.0", message="Syslog changed to JSON format. Use the send_syslog_event function instead.")
@task_event(
    {
        405: EventConfigurationError(
            f"Cannot send {Events.CEF.event_name} as syslog is either disabled or not configured in the PPA appliance."
        )
    }
)
def send_cef_event(
    *,
    message: str,
    level: Optional[str] = CEFEventLevels.INFO,
    reason: Optional[str] = None,
    fields: Optional[dict] = None,
) -> None:
    """
    This function is retired in PPA appliance version 2.5.0!

    Send a CEF syslog event using the PPA syslog configuration.

    :param message: Event message.
    :param level: Event level. Defaults to `INFO`. Valid levels are listed below.
    :param reason: The text to show in the PPA task interface when the CEF event is sent. If it is not supplied, no
        text will be shown.
    :param fields: A dictionary containing any extra CEF fields to supply in the event.

    :raises exceptions.EventConfigurationError: If either there is no PPA appliance syslog configuration, or the PPA
        appliance syslog configuration is disabled.

    Valid event levels:

        * events.CEFEventLevels.INFO
        * events.CEFEventLevels.PANIC
        * events.CEFEventLevels.FATAL
        * events.CEFEventLevels.ERROR
        * events.CEFEventLevels.WARNING
        * events.CEFEventLevels.DEBUG
        * events.CEFEventLevels.TRACE

    Example:

        .. code-block:: python

            events.send_cef_event(
                subject="An Active Directory user was moved by PPA",
                reason="Sending CEF event",
                fields={
                    "cs1Label": "task_name",
                    "cs1": "Move Active Directory User"
                }
            )
    """
    return event("message/syslog", reason, message=message, level=level, fields=fields)


@min_ppa_version("2.5.0")
@task_event(
    {
        405: EventConfigurationError(
            "Cannot send syslog event as syslog is either disabled or not configured in the PPA appliance."
        )
    }
)
def send_syslog_event(
    *, message: str, level: Optional[str] = "info", reason: Optional[str] = None, details: Optional[dict] = None,
) -> None:
    """ Send a JSON syslog event using the PPA syslog configuration.

    :param message: Event message.
    :param reason: The text to show in the PPA task interface when the syslog event is sent. If it is not supplied, no
        text will be shown.
    :param level: Event level. Defaults to `info`. Valid levels are listed below.
    :param details: A dictionary containing custom event information.

    :raises exceptions.EventConfigurationError: If either there is no PPA appliance syslog configuration, or the PPA
        appliance syslog configuration is disabled.

    Valid event levels are:

        * info
        * panic
        * fatal
        * error
        * warning
        * debug
        * trace

    Example:

        .. code-block:: python

            events.send_syslog_event(
                message="An Active Directory user password was changed by PPA",
                reason="Sending syslog event",
                details={
                    "active_directory_user": f"{active_directory_user.sAMAccountName}",
                    "change_at_next_logon": True,
                    "unlocked": False
                }
            )
    """
    # For now 'details' is sent to the appliance as 'fields' - https://github.com/Osirium/ppa/issues/445
    return event("message/syslog", reason, message=message, level=level, fields=details)


def create_wait_hook() -> Tuple[str, str]:
    """ Generate a wait hook URL and secret. This should be used when an approval
    process is required. The generated URL can be sent to an approver from inside
    the task, and the secret is passed to the `events.wait` function to start the
    approval process.

    :returns: A tuple of strings containing the URL and secret.

    Example:

        .. code-block:: python

            url, secret = events.create_wait_hook()

    """
    secret = base64.urlsafe_b64encode(os.urandom(24)).decode("utf-8")
    return (f"{os.environ['PPA_WAIT_HOOK']}/{secret}", secret)


@task_event()
def wait(
    text: str, *, secret: str, element_id: Optional[str] = None, template: Optional[str] = None
) -> Dict[str, Union[str, list, dict]]:
    """ Block any task activity until a specific wait hook URL has been triggered.
    This function should be called after generating a url and secret with
    `events.create_wait_hook`.

    Once the URL is clicked by any party, they will be navigated to the PPA Appliance interface
    and asked to authenticate before the response is processed.

    Multiple actions can be triggered by appending parameters to the generated URL. See
    the code examples below for more information.

    :param text: The text to display inside the task when the wait starts.
    :param secret: A secret generated by `create_wait_hook`.
    :param element_id: The ID of the UI element. See :ref:`getting_started:Input & Output IDs`.
    :param template: A mustache template to display to the responder.

    :returns: A dictionary of response information. The `username` key contains the
        username of the responder. The `values` key contains a dictionary of URL
        parameter names and values.

    Simple example with no URL parameters:

        .. code-block:: python

            # generate a simple URL with no parameters
            url, secret = events.create_wait_hook()
            # now wait for the URL to be clicked, and for the responder
            # to authenticate
            events.wait("Waiting for the URL to be clicked...", secret=secret)

    Example with `approve` and `reject` URL parameters:

        .. code-block:: python

            from distutils.util import strtobool

            # generate a base URL with no parameters
            url, secret = events.create_wait_hook()

            # create approve and reject URLs
            approve_url = f"{url}?approve=true"
            reject_url = f"{url}?approve=false"

            # now deliver these URLs to wherever you like, this function is hypothetical
            mail.send(approver, f"{approve_url}, {reject_url}")

            # wait for either URL to be clicked, and for the responder to authenticate
            # Example response: {'username': 'domain\\line.manager', 'values': {'approve': ['true']}}
            response = events.wait(
                "Waiting for approval response...",
                secret=secret,
                template='''
                <%#values.success%>
                # Success!
                Responder: <%username%>
                <%/values.success%>
                <%^values.success%>
                # Failed!
                Responder: <%username%>
                <%/values.success%>
                '''
            )
            approved = strtobool(response["values"]["approve"])
            responder = response["username"]

            if approve:
                ui.output_info(f"Request was approved by {responder}.")
            else:
                ui.output_error(f"Request was rejected by {responder}.")

    """
    return event("hooks/wait", text, secret=secret, id=element_id, template=template)
