import logging
import functools
import json
import contextlib

from typing import Callable, List, Dict, Any, Optional, Tuple

import spnego

from ppa import ui
from ppa.exceptions import PPAVersionError
from windows_server import client, services, exceptions as sdk_exceptions
from windows_server.exceptions import CertificateError, NetworkError, MissingKerberosConfig

from .records import WindowsServer


logger = logging.getLogger(__name__)


@contextlib.contextmanager
def error_handler(errors: Optional[List[Tuple[str, str, bool]]] = None, default: Optional[str] = None):
    errors = [] if not errors else errors

    try:
        yield
    except sdk_exceptions.PowershellError as e:
        logger.exception(e)
        powershell_error = str(e)

        # Set defaults
        message = default or f"Powershell Error: {powershell_error}"
        raise_exception = True

        for error, custom_message, fail in errors:
            if error.lower() in powershell_error.lower():
                # Overwrite defaults if a matching custom error was supplied
                message = custom_message
                raise_exception = fail
                break

        ui.output_error(message)
        if raise_exception:
            raise sdk_exceptions.PowershellError(message)


def json_load(output: str) -> List[Dict[str, Any]]:
    try:
        return json.loads(output)
    except json.JSONDecodeError as e:
        logger.error(f"output: {output}")
        ui.output_error(
            "Failed to parse the Powershell output as it is not valid JSON. Please view the logs for more information."
        )
        raise e


def is_domain_controller(windows_server: WindowsServer) -> bool:
    # ProductType
    #
    # Data type: uint32
    #
    # Access type: Read-only
    #
    # Additional system information.
    #
    # Work Station (1)
    # Domain Controller (2)
    # Server (3)
    output = client.run_powershell(
        windows_server,
        script="""$osInfo = Get-CimInstance -ClassName Win32_OperatingSystem
$osInfo.ProductType""",
    )
    try:
        return {1: False, 2: True, 3: False}[int(output)]
    except KeyError:
        logger.error(f"unknown product type {output}")
        message = "Couldn't establish whether the server is a Domain Controller. Please view the logs for more information."
        ui.output_error(message)
        raise KeyError(message)
    except Exception as e:
        logger.error(f"failed to establish product type from windows server output: '{output}'")
        raise e


def requires(service_name: str):
    def _service_checker(func):
        def _execute(windows_server, *args, **kwargs):
            logger.info(f"checking the supplied server has required service {service_name}")
            if not services.get_by_name(windows_server, name=service_name):
                message = f"The action failed as the server is not running the {service_name} service."
                ui.output_error(message)
                raise Exception(message)
            return func(windows_server, *args, **kwargs)

        return _execute

    return _service_checker


def winrm(func: Callable) -> Callable:
    @functools.wraps(func)
    def _winrm_command(windows_server: WindowsServer, *args, **kwargs):
        # These apply to both auth methods.
        connection_args = dict(
            host=windows_server["address"],
            use_ssl=windows_server.get("use_ssl", True),
            validate_cert=windows_server.get("validate_cert", True),
        )

        for timeout_key in {"read_timeout", "operation_timeout"}:
            # Don't supply these with None value, they must be omitted completely if
            # not supplied, otherwise the defaults will not be used by the SDK.
            if timeout_value := windows_server.get(timeout_key):
                connection_args[timeout_key] = timeout_value

        logger.debug(f"base connection arguments: {connection_args}")

        ntlm_auth_keys = {"username", "password"}
        if ntlm_keys := [key for key in ntlm_auth_keys if key in windows_server]:
            if len(ntlm_keys) == 1:
                message = f"Invalid connection details. Received a {ntlm_keys[0]} without a {[k for k in ntlm_auth_keys if k not in ntlm_keys][0]}."
                ui.output_error(message)
                raise KeyError(message)
            logger.info("plugin will use ntlm authentication")
            initializer = client.initialize_ntlm
            # These only apply to the NTLM auth method
            connection_args["username"] = windows_server["username"]
            connection_args["password"] = windows_server["password"]
        else:
            logger.info("plugin will use kerberos authentication")
            initializer = client.initialize_kerberos

        try:
            with initializer(**connection_args)() as open_connection:
                return func(open_connection, *args, **kwargs)

        except PPAVersionError as e:
            ui.output_error("Kerberos authentication requires PPA v2.7.0 or newer.")
            raise e

        except (CertificateError, NetworkError) as e:
            ui.output_error(
                f"Connection failed due to a {'certificate' if isinstance(e, CertificateError) else 'network'} "
                "error. Please view the logs for more information."
            )
            raise e

        except MissingKerberosConfig as e:
            ui.output_error("The task attempted Kerberos authentication, but no configuration was found.")
            ui.output_error(
                "If Kerberos is required, make sure it is enabled for this task. "
                "Otherwise supply a username & password for authentication."
            )
            raise e

        except spnego.exceptions.SpnegoError as e:
            ui.output_error("Failed to negotiate Kerberos authentication. Please view the logs for more information.")
            raise e

    return _winrm_command
