"""
Domain

This module contains actions for related to Active Directory domains.

!!! warning "Domain Servers"

    The actions in this module will only work on domain joined servers.

    If an action requires a Powershell module to be installed, its documentation will let you know.
"""
import re
import logging

from ppa import ui
from windows_server import client

from .records import WindowsServer
from ._wrappers import winrm, error_handler


logger = logging.getLogger(__name__)


@winrm
def get_netbios_domain_name(windows_server: WindowsServer) -> str:
    r"""Get the NetBIOS name of the Active Directory domain the server is joined to.

The NetBIOS domain name is the short domain used in the _domain\username_ format.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

??? success "Output"

    The NetBIOS domain name.

??? example

    ```YAML linenums="1"
    - windows_server.domain.get_netbios_domain_name:
      load:
        windows_server: windows_server
      save: netbios_domain_name
    ```
"""

    # 2012-DOMAIN    <00>  GROUP       Registered
    regex = re.compile(r"(.*)\s+<00>\s+GROUP\s+Registered")
    with error_handler(default="Failed to get NetBIOS domain name on the server."):
        output = client.run_powershell(windows_server, script="nbtstat -n")

    logger.debug(f"nbtstat output: {output}")
    for line in output.split("\n"):
        if match := re.search(regex, line):
            return match.group(1).strip()

    ui.output_error("Could not identify NetBIOS domain name on the server. Please view the logs for more information.")
    raise Exception(r"No NetBIOS domain name could not be found (no record matching '<00> GROUP Registered').")
