from typing import TypedDict, Optional

from datacaster.classes import CastDataClass


RDP_SESSION_STATES = {"disc": "Disconnected", "listen": "Listening", "conn": "Connected", "active": "Active"}


class WindowsServer(TypedDict):
    r"""
- Contains remote Powershell connection details for a Windows Server

- It is used as an _input_ to many of the actions in this plugin

- The example below shows the dictionary structure in YAML

- See the ___required___ and _optional_ keys underneath the example

!!! tip "Authentication Methods"

    PPA v2.7.0 introduces support for Kerberos authentication, on top of the existing NTLM support.

    If a **username** & **password** are supplied, NTLM authentication will be attempted.

    To use Kerberos authentication, just omit the **username** & **password**.

    Kerberos support requires plugin version **2.0.0** and newer.

!!! warning "Supplying Credentials"

    You should ___always___ use a PPA Vault integration to provide credentials to a plugin action.

!!! important "Authenticating Using Kerberos"

    When authenticating to a Windows server using Kerberos, the server must have the following Service Principal Names set:

    ```
        WSMAN/{hostname}
        WSMAN/{FQDN}
    ```

    For more information regarding Service Principal Names, click [here](https://docs.microsoft.com/en-us/windows/win32/ad/service-principal-names)

!!! example "Kerberos Authentication"

    Requires PPA v2.7.0 or newer, & plugin version 2.0.0 or newer.

    ```yaml linenums="1"
    windows_server:
      address: windows-server.internal.net
    ```

!!! example "NTLM Authentication"

    ```yaml linenums="1"
    windows_server:
      address: windows-server.internal.net
      username: example_username
      password: example_password
    ```

!!! danger "Required Keys"

    __address__: The target IP or DNS address.

!!! info "Optional Keys"

    __username__: Username for authentication (if using NTLM authentication).

    __password__: Password for authentication (if using NTLM authentication).

    __validate_cert__: Set to _false_ to skip certificate validation (defaults to _true_).

    __use_ssl__: Set to _false_ to connect over HTTP (defaults to _true_ using HTTPS).

    __read_timeout__: The timeout in seconds for receiving a response from the server (defaults to _30_).

    __operation_timeout__: The timeout in seconds for each remote Powershell operation (defaults to _20_.
    """
    address: str
    username: Optional[str] = None
    password: Optional[str] = None
    validate_cert: Optional[bool] = True
    use_ssl: Optional[bool] = True
    read_timeout: Optional[int] = None
    operation_timeout: Optional[int] = None


class RDPSessionData(CastDataClass):
    id: Optional[int]
    username: str
    session_name: str
    state: str
    __class_config__ = {
        "rename_fields": {"sessionname": "session_name"},
        "cast_functions": {
            "fields": {"id": lambda x: int(x) if x else None, "state": lambda state: RDP_SESSION_STATES.get(state.lower(), state)}
        },
        "always_cast": ["state"],
    }


class RDPSession(TypedDict):
    r"""
- Contains RDP session details

- The example below shows the dictionary structure in YAML

!!! example

    ```yaml linenums="1"
    rdp_session:
      username: domain\username
      session_name: rdp-tcp#50
      state: Active
      id: 2
    ```

???+ info "All RDPSession Keys"

    __id__: The RDP session ID on the Windows Server.

    __username__: The username the RDP session was authenticated with.

    __session_name__: The name of the RDP session.

    __state__: The state of the RDP session (`Disconnected`, `Connected`, `Active`, `Listening`).
    """
    __annotations__ = RDPSessionData.__annotations__


class ServiceData(CastDataClass):
    name: str
    display_name: str
    state: str
    status: str
    start_name: str
    description: str
    __class_config__ = {
        "rename_fields": {"displayname": "display_name", "startname": "start_name"},
    }


class Service(TypedDict):
    r"""
- Contains Windows service details

- The example below shows the dictionary structure in YAML

!!! example

    ```yaml linenums="1"
    name: Dhcp
    display_name: DHCP Client
    state: Running
    status: OK
    start_name: NT Authority\\LocalService
    description: >
      Registers and updates IP addresses and DNS records for this computer.
      If this service is stopped, this computer will not receive dynamic IP addresses
      and DNS updates. If this service is disabled, any services that explicitly depend
      on it will fail to start.
    ```

???+ info "All Service Keys"

    __name__: The name of the service.

    __display_name__: The display name for the service.

    __state__: The state of the service.

    __status__: The service status.

    __start_name__: The account the service runs under.

    __description__: The service description.
    """
    __annotations__ = ServiceData.__annotations__


class UserData(CastDataClass):
    name: str
    full_name: str
    description: str
    disabled: bool
    locked: bool
    password_changeable: bool
    password_expires: bool
    password_required: bool
    sid: str
    __class_config__ = {
        "rename_fields": {
            "Name": "name",
            "FullName": "full_name",
            "Description": "description",
            "Disabled": "disabled",
            "Lockout": "locked",
            "PasswordChangeable": "password_changeable",
            "PasswordExpires": "password_expires",
            "PasswordRequired": "password_required",
            "SID": "sid",
        },
    }


class User(TypedDict):
    """
- Contains Windows user account details

- The example below shows the dictionary structure in YAML

!!! example

    ```YAML linenums="1"
    name: Administrator
    full_name: ''
    description: Built-in account for administering the computer/domain
    disabled: false
    locked: false
    password_changeable: true
    password_expires: true
    password_required: true
    sid: S-1-5-21-2746729022-1139076888-119054876-500
    ```

!!! info "User Keys"

    __name__: The name of the user.

    __full_name__: The user's full name.

    __description__: The user description.

    __disabled__: Indicates whether the user is disabled.

    __locked__: Indicates whether the user is locked.

    __password_changeable__: Indicates whether the user can change their own password.

    __password_expires__: Indicates whether the user's password will expire.

    __password_required__: Indicates whether a password is required to log on as this user.

    __sid__: The user account object SID.
    """

    __annotations__ = UserData.__annotations__


class GroupData(CastDataClass):
    name: str
    description: str
    sid: str
    __class_config__ = {
        "rename_fields": {"Name": "name", "Description": "description", "SID": "sid"},
    }


class Group(TypedDict):
    """
- Contains Windows group details

!!! example

    ```YAML linenums="1"
    name: Administrators
    description: Administrators have complete and unrestricted access to the computer/domain
    sid: S-1-5-21-2746729022-1139076888-119054876-500
    ```

!!! info "Group Keys"

    __name__: The name of the group.

    __description__: The group description.

    __sid__: The group object SID.
    """

    __annotations__ = GroupData.__annotations__


class DNSZoneData(CastDataClass):
    """
    ZoneName                            ZoneType        IsAutoCreated   IsDsIntegrated  IsReverseLookupZone  IsSigned
    --------                            --------        -------------   --------------  -------------------  --------
    _msdcs.2012.osirium.tom             Primary         False           True            False                False
    0.in-addr.arpa                      Primary         True            False           True                 False
    127.in-addr.arpa                    Primary         True            False           True                 False
    2012.osirium.tom                    Primary         False           True            False                False
    255.in-addr.arpa                    Primary         True            False           True                 False
    TrustAnchors                        Primary         False           True            False                False
    """

    name: str
    type: str
    is_auto_created: bool
    is_ds_integrated: bool
    is_reverse_lookup_zone: bool
    is_signed: bool

    __class_config__ = {
        "rename_fields": {
            "ZoneName": "name",
            "ZoneType": "type",
            "IsAutoCreated": "is_auto_created",
            "IsDsIntegrated": "is_ds_integrated",
            "IsReverseLookupZone": "is_reverse_lookup_zone",
            "IsSigned": "is_signed",
        }
    }


class RecordTTL(CastDataClass):
    days: int
    hours: int
    minutes: int
    seconds: int

    __class_config__ = {"rename_fields": {"Days": "days", "Hours": "hours", "Minutes": "minutes", "Seconds": "seconds"}}


class DNSZone(TypedDict):
    """
- Contains Windows DNS zone details

!!! example

    ```YAML linenums="1"
    name: internal.net
    type: Primary
    is_auto_created: false
    is_ds_integrated: true
    is_reverse_lookup_zone: false
    is_signed: false
    ```

!!! info "DNSZone Keys"

    __name__: The name of the zone.

    __type__: The zone type.

    __is_auto_created__: Indicates whether the zone was created automatically.

    __is_ds_integrated__: Indicates whether the zone is integrated with Active Directory.

    __is_reverse_lookup_zone__: Indicates whether the zone is a reverse lookup zone.

    __is_signed__: Indicates whether the zone is signed.
    """

    __annotations__ = DNSZoneData.__annotations__


class DNSRecordData(CastDataClass):
    """
    HostName                  RecordType Timestamp            TimeToLive      RecordData
    --------                  ---------- ---------            ----------      ----------
    @                         A          21/12/2020 15:00:00  00:10:00        10.9.206.80
    @                         NS         0                    01:00:00        dc1.2012.osirium.tom.
    @                         SOA        0                    01:00:00        [2254][dc1.2012.osirium.tom.][hostmaster.2...
    _msdcs                    NS         0                    01:00:00        win-hhum3srbe2s.2012.osirium.tom.
    _gc._tcp.Default-First... SRV        21/12/2020 15:00:00  00:10:00        [0][100][3268][dc1.2012.osirium.tom.]
    _kerberos._tcp.Default... SRV        21/12/2020 15:00:00  00:10:00        [0][100][88][dc1.2012.osirium.tom.]
    _ldap._tcp.Default-Fir... SRV        21/12/2020 15:00:00  00:10:00        [0][100][389][dc1.2012.osirium.tom.]
    _gc._tcp                  SRV        21/12/2020 15:00:00  00:10:00        [0][100][3268][dc1.2012.osirium.tom.]
    _kerberos._tcp            SRV        21/12/2020 15:00:00  00:10:00        [0][100][88][dc1.2012.osirium.tom.]
    _kpasswd._tcp             SRV        21/12/2020 15:00:00  00:10:00        [0][100][464][dc1.2012.osirium.tom.]
    _ldap._tcp                SRV        21/12/2020 15:00:00  00:10:00        [0][100][389][dc1.2012.osirium.tom.]
    _kerberos._udp            SRV        21/12/2020 15:00:00  00:10:00        [0][100][88][dc1.2012.osirium.tom.]
    _kpasswd._udp             SRV        21/12/2020 15:00:00  00:10:00        [0][100][464][dc1.2012.osirium.tom.]
    dc1                       A          0                    01:00:00        10.9.206.80
    """

    hostname: str
    type: str
    time_to_live: RecordTTL
    data: str

    __class_config__ = {
        "rename_fields": {"HostName": "hostname", "RecordType": "type", "TimeToLive": "time_to_live", "RecordData": "data"},
        "cast_functions": {
            "fields": {
                "time_to_live": lambda ttl: RecordTTL(**ttl).__dict__,
                "data": lambda data: data.get("CimInstanceProperties") or "",
            },
        },
        "always_cast": ["data"],
    }


class DNSRecord(TypedDict):
    """
- Contains Windows DNS record details

!!! example "Example - A Record"

    ```YAML linenums="1"
    hostname: iis-001
    type: A
    time_to_live:
      days: 0
      hours: 1
      minutes: 0
      seconds: 0
    data: IPv4Address = "10.9.10.10"
    ```

!!! example "Example - CName Record"

    ```YAML linenums="1"
    hostname: web-server
    type: CNAME
    time_to_live:
      days: 0
      hours: 1
      minutes: 0
      seconds: 0
    data: HostNameAlias = "iis-001.internal.net."
    ```

!!! example "Example - PTR Record"

    ```YAML linenums="1"
    hostname: 10.9.10.10
    type: PTR
    time_to_live:
      days: 0
      hours: 1
      minutes: 0
      seconds: 0
    data: PtrDomainName = "iis-001.internal.net"
    ```

!!! info "DNSRecord Keys"

    __hostname__: The record hostname.

    __type__: The record type (_A_, _CNAME_, or _PTR_).

    __time_to_live__: Dictionary containing the TTL in _days_, _hours_, _minutes_, & _seconds_.

    __data__: Record-specific data (see examples above for more).
    """

    __annotations__ = DNSZoneData.__annotations__
