"""
Users

This module contains actions for auditing _local_ users on a Windows Server.
"""
import logging

from typing import List, Dict, Any, Union, Optional

from ppa import ui
from windows_server import client
from playbook_tools import tables

from .records import WindowsServer, User, UserData
from ._wrappers import winrm, is_domain_controller, json_load


logger = logging.getLogger(__name__)

USER_TABLE_HEADER = ["Name", "Full Name", "Description", "Disabled", "Locked"]
USER_TABLE_FIELDS = ["name", "full_name", "description", "disabled", "locked"]


def _parse_users(json_output: List[Dict[str, Any]]) -> List[User]:
    return [UserData(**user).__dict__ for user in json_output]


@winrm
def get_all(windows_server: WindowsServer) -> Union[List[User], list]:
    """Get all local users from a Windows Server.

!!! question "Minimum Plugin Version: 2.0.0"

!!! danger "Domain Controllers"

    Running this action against a Domain Controller will return an empty list.

??? info "Input"

    - __windows_server:__ a [WindowsServer](../inputs_outputs/#windowsserver) dictionary

??? success "Output"

    A list of [User](../inputs_outputs/#user) dictionaries.

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users
    ```
"""
    if is_domain_controller(windows_server):
        return []

    output = client.run_powershell(
        windows_server, script="Get-WmiObject -Class Win32_UserAccount -Filter \"LocalAccount='True'\" | ConvertTo-JSON"
    )

    if not output:
        message = "Failed to get users as the response from the server was empty."
        ui.output_error(message)
        raise ValueError(f"{message} Output: {repr(output)}.")

    return _parse_users(json_load(output))


def output_table(text: str, users: Union[User, List[User]]) -> None:
    """Display a list of users in a table.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 1.4.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ a list of [User](../inputs_outputs/#user) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - windows_server.users.output_table:
        text: All Local Users
      load:
        users: all_users
    ```
    """
    ui.output_table(
        text,
        header=USER_TABLE_HEADER,
        data=[
            [str(record.get(field)) if record.get(field) is not None else "" for field in USER_TABLE_FIELDS] for record in users
        ],
    )


def display(text: str, users: Union[User, List[User]]) -> None:
    """Display users in a table.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - windows_server.users.display:
        text: All Local Users
      load:
        users: all_users
    ```
    """
    tables.output_table(text, header=USER_TABLE_HEADER, extract=USER_TABLE_FIELDS, records=users)


def display_tabbed(text: str, tabs: Dict[str, Union[User, List[User]]]) -> None:
    """Display users in a table with multiple tabs.

This action can be used to display user lists from multiple servers.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is any number of [Users](../inputs_outputs/#user)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - set:
        name: tabs
        value: {}

    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - ppa_tools.dictionaries.insert:
        name: My Server
      load:
        value: all_users
        dictionary: tabs
      save: tabs

    - windows_server.users.display_tabbed:
        text: All Local Users
      load:
        tabs: tabs
    ```
    """
    tables.output_tabbed_table(
        text, header=USER_TABLE_HEADER, extract=USER_TABLE_FIELDS, tabs=tabs,
    )


def output_tabbed_table(text: str, tabs: Dict[str, List[User]]) -> None:
    """Display a users table with multiple tabs.

This action can be used to display user lists from multiple servers.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display_tabbed](#display_tabbed).

!!! question "Minimum Plugin Version: 1.5.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is a list of [User](../inputs_outputs/#user) dictionaries

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - set:
        name: tabbed_table
        value: {}

    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: all_users

    - ppa_tools.dictionaries.insert:
        name: My Server
      load:
        value: all_users
        dictionary: tabbed_table
      save: tabbed_table

    - windows_server.users.output_tabbed_table:
        text: All Local Users
      load:
        tabs: tabbed_table
    ```
    """
    ui.output_table(
        text,
        header=USER_TABLE_HEADER,
        data={
            name: [
                [str(record.get(field)) if record.get(field) is not None else "" for field in USER_TABLE_FIELDS]
                for record in records
            ]
            for name, records in tabs.items()
        },
    )


def select(
    text: str, users: Union[User, List[User]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> tables.TableSelection:
    """Display users in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Users](../inputs_outputs/#user)

    - **all**: a list of selected [Users](../inputs_outputs/#user)

    - **first**: the first selected [User](../inputs_outputs/#user)

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: users

    - windows_server.users.select:
        text: Select Users
      load:
        users: users
      save: selection
    ```
    """
    return tables.input_table(
        text, header=USER_TABLE_HEADER, extract=USER_TABLE_FIELDS, records=users, minimum=minimum, maximum=maximum
    )


def select_one(text: str, users: Union[User, List[User]]) -> User:
    """Display users in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Full Name
- Description
- Disabled
- Locked

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

??? success "Output"

    A single [User](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - windows_server.users.get_all:
      load:
        windows_server: windows_server
      save: users

    - windows_server.users.select_one:
        text: Select a User
      load:
        users: users
      save: user
    ```
    """
    return select(text, users, minimum=1, maximum=1)["first"]
